import logging
import os
import sys
import time

import numpy as np
import pandas as pd
import snowflake.connector
import xgboost as xgb
from flask import Flask
from flask import make_response
from flask import request
from sklearn import preprocessing
from snowflake.snowpark import Session
from snowflake.snowpark.functions import col

SERVICE_HOST = os.getenv("SERVER_HOST", "0.0.0.0")
SERVICE_PORT = os.getenv("SERVER_PORT", 8080)


def get_logger(logger_name):
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)
    handler = logging.StreamHandler(sys.stdout)  # Changed to stdout for better Docker compatibility
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(
        logging.Formatter(
            "%(name)s [%(asctime)s] [%(levelname)s] %(message)s"))
    logger.addHandler(handler)
    return logger


def handle_request(message, logger):
    logger.debug(f"Received request: {message}")

    if message is None or not message["data"]:
        logger.info("Received empty message")
        return None

    input_rows = message["data"]
    logger.info(f"Received {len(input_rows)} rows")
    return input_rows


def make_json_response(data):
    response = make_response({"data": data})
    response.headers["Content-type"] = "application/json"
    logger.debug(f"Prepared response: {response.json}")
    return response


# In-memory model
model = None
ohe = None

logger = get_logger("lal-service")

app = Flask(__name__)


@app.get("/healthcheck")
def readiness_probe():
    return "I'm ready!"


@app.post("/train")
def train():
    """
    Train the in-memory model.
    Takes table and column names as input (not the actual rows) and queries the data directly.
    """
    global model
    global ohe

    # Input format: https://docs.snowflake.com/en/developer-guide/snowpark-container-services/working-with-services#data-exchange-format
    input_rows = handle_request(request.json, logger)
    if len(input_rows) != 1:
        return make_json_response([[0, {"error": "Expected exactly one input"}]])
    req = input_rows[0]
    provider_table = req[1]  # source_table[0]
    provider_join_col = req[2]  # provider_join_col
    consumer_table = req[3]  # my_table[0]
    consumer_join_col = req[4]  # consumer_join_col
    dimensions = req[5]  # dimensions
    filter_clause = req[6]  # filter_clause

    start_time = time.perf_counter()
    session = snow_session()

    provider_dimensions = []
    for d in dimensions:
        d = d.replace(' ', '')
        provider_dimensions.append(d)
    dimensions_str = ",".join(provider_dimensions)
    query = f"""
        WITH
        features AS (
            SELECT
                {provider_join_col} as join_col,
                {dimensions_str}
            FROM
                identifier('{provider_table}') as p
        ),
        labels AS (
            SELECT
                {consumer_join_col} as join_col,
                {filter_clause} as label_value
            FROM
                identifier('{consumer_table}') as c
        ),
        training_data AS (
            SELECT
                f.*,
                l.label_value
            FROM
                features f, labels l
            WHERE
                f.join_col = l.join_col
        )
        SELECT * FROM training_data
    """

    try:
        joined_df = session.sql(query).drop(col("join_col")).to_pandas()
        labels = joined_df.filter(["LABEL_VALUE"], axis=1)
        features = joined_df.drop(columns=["LABEL_VALUE"])

        categorical = features.select_dtypes(include=[object])
        ohe = preprocessing.OneHotEncoder(handle_unknown="ignore")
        categorical_ohe = ohe.fit_transform(categorical)
        non_categorical = features.select_dtypes(exclude=[object])
        X_ready = np.concatenate((categorical_ohe.toarray(), non_categorical.to_numpy()), axis=1)
        xg_train = xgb.DMatrix(X_ready, label=labels)
        params = {"objective": "binary:logistic", "max_depth": 3, "eta": 1, "nthread": 1, "eval_metric": "auc"}
        evallist = [(xg_train, "train")]
        evals_result = {}

        model = xgb.train(
            params, xg_train, 10, evallist, evals_result=evals_result
        )

        auc = np.max(evals_result["train"]["auc"])
        error = 1 - np.max(evals_result["train"]["auc"])

        end_time = time.perf_counter()
        execution_time = end_time - start_time
        response = make_json_response([[0, {"auc": auc, "error": error, "execution_time": round(execution_time, 2)}]])
        return response
    except Exception as e:
        response = make_json_response([[0, {"error": str(e)}]])
        return response


@app.post("/score")
def score():
    """
    Perform inference using the trained in-memory model.
    Takes table and column names as input (not the actual rows) and queries the data directly.
    """
    global model
    global ohe

    if not model:
        return make_json_response([[0, {"error": "Model has not been trained"}]])

    # Input format: https://docs.snowflake.com/en/developer-guide/snowpark-container-services/working-with-services#data-exchange-format
    input_rows = handle_request(request.json, logger)
    if len(input_rows) != 1:
        return make_json_response([[0, {"error": "Expected exactly one input"}]])
    req = input_rows[0]
    provider_table = req[1]
    provider_join_col = req[2]
    consumer_table = req[3]
    consumer_join_col = req[4]
    dimensions = req[5]

    start_time = time.perf_counter()
    session = snow_session()

    try:
        provider_dimensions = []
        for d in dimensions:
            d = d.replace(' ', '')
            provider_dimensions.append(d)
        dimensions_str = ','.join(provider_dimensions)
        query = f"""
            SELECT {provider_join_col} as id, {dimensions_str}
                 FROM identifier('{provider_table}') as p
                 WHERE {provider_join_col} NOT IN (SELECT {consumer_join_col} FROM identifier('{consumer_table}') as c)
        """
        exclusive_df = session.sql(query).to_pandas()

        training_input_identities = exclusive_df.filter(["ID"], axis=1)
        training_input_features = exclusive_df.drop(columns=["ID"])

        Y1 = training_input_features.select_dtypes(include=[object])
        Y2 = ohe.transform(Y1)

        # get the rest of the features and add them to the binary features
        Y3 = training_input_features.select_dtypes(exclude=[object])
        Y_pred = np.concatenate((Y2.toarray(), Y3.to_numpy()), axis=1)

        dscore = xgb.DMatrix(Y_pred)
        pred = model.predict(dscore)

        retval = list(zip(np.array(training_input_identities), list(map(str, pred))))
        ids = []
        scores = []
        for ret in retval:
            ids.append(ret[0][0])
            scores.append(ret[1])

        df = pd.DataFrame({
            "ID": ids,
            "SCORE": scores
        })
        table = "LOOKALIKE_RESULTS"
        session.write_pandas(df, table, schema="CLEANROOM", auto_create_table=True, overwrite=True)

        end_time = time.perf_counter()
        execution_time = end_time - start_time
        response = make_json_response([[0, {"results_table": table, "size": len(ids), "execution_time": round(execution_time, 2)}]])
        return response
    except Exception as e:
        response = make_json_response([[0, str(e)]])
        return response


@app.post("/scorebatch")
def score_batch():
    """
    Perform inference using the trained in-memory model.
    Takes rows of data as input and support batching.
    """
    global model
    global ohe

    if not model:
        return make_json_response([[0, {"error": "Model has not been trained"}]])

    input_rows = handle_request(request.json, logger)
    if not input_rows:
        return {}

    start_time = time.perf_counter()

    try:
        training_input_features = pd.DataFrame([r[2] for r in input_rows])

        Y1 = training_input_features.select_dtypes(include=[object])
        Y2 = ohe.transform(Y1)

        # get the rest of the features and add them to the binary features
        Y3 = training_input_features.select_dtypes(exclude=[object])
        Y_pred = np.concatenate((Y2.toarray(), Y3.to_numpy()), axis=1)

        dscore = xgb.DMatrix(Y_pred)
        pred = model.predict(dscore)

        retval = [
            [row[0], {"id": row[1], "score": str(score)}]
            for row, score in zip(input_rows, pred)
        ]

        end_time = time.perf_counter()
        execution_time = end_time - start_time
        response = make_json_response(retval)
        return response
    except Exception as e:
        response = make_json_response([[0, {"error": str(e)}]])
        return response


def snow_connection() -> snowflake.connector.SnowflakeConnection:
    if os.path.isfile("/snowflake/session/token"):
        creds = {
            "host": os.getenv("SNOWFLAKE_HOST"),
            "port": os.getenv("SNOWFLAKE_PORT"),
            "protocol": "https",
            "account": os.getenv("SNOWFLAKE_ACCOUNT"),
            "authenticator": "oauth",
            "token": open("/snowflake/session/token", "r").read(),
            "warehouse": os.getenv("SNOWFLAKE_WAREHOUSE"),
            "database": os.getenv("SNOWFLAKE_DATABASE"),
            "schema": os.getenv("SNOWFLAKE_SCHEMA"),
            "client_session_keep_alive": True
        }
    else:
        creds = {
            "account": os.getenv("SNOWFLAKE_ACCOUNT"),
            "user": os.getenv("SNOWFLAKE_USER"),
            "password": os.getenv("SNOWFLAKE_PASSWORD"),
            "warehouse": os.getenv("SNOWFLAKE_WAREHOUSE"),
            "database": os.getenv("SNOWFLAKE_DATABASE"),
            "schema": os.getenv("SNOWFLAKE_SCHEMA"),
            "client_session_keep_alive": True
        }

    connection = snowflake.connector.connect(**creds)
    return connection


def snow_session() -> Session:
    return Session.builder.configs({"connection": snow_connection()}).create()


if __name__ == "__main__":
    app.run(host=SERVICE_HOST, port=SERVICE_PORT)
